class NotAllowedInstantiation(Exception):
    pass


class InvalidMimeoUtil(Exception):
    pass


class NotASpecialField(Exception):
    pass


class InvalidSpecialFieldValue(Exception):
    pass


class CountryNotFound(Exception):
    pass


class OutOfStock(Exception):
    pass
