class UnsupportedOutputFormat(Exception):
    pass


class UnsupportedOutputDirection(Exception):
    pass


class UnsupportedAuthMethod(Exception):
    pass


class UnsupportedRequestMethod(Exception):
    pass


class MissingRequiredProperty(Exception):
    pass


class InvalidIndent(Exception):
    pass


class InvalidVars(Exception):
    pass


class IncorrectMimeoModel(Exception):
    pass


class IncorrectMimeoTemplate(Exception):
    pass


class IncorrectMimeoConfig(Exception):
    pass

class ResourceNotFound(Exception):
    pass
